// -*- C++ -*-

/**
 * @file alignedmmap.h
 * @author Emery Berger <http://www.cs.umass.edu/~emery>
 */


#ifndef _ALIGNEDMMAP_H_
#define _ALIGNEDMMAP_H_

#if defined(_WIN32)
//#error "You should not include this header on Windows."
#endif

#include "sassert.h"
#include "stlallocator.h"
#include "freelistheap.h"
#include "mmapwrapper.h"
#include "bumpalloc.h"
#include "exactlyone.h"
#include "mmapalloc.h"
#include "myhashmap.h"

using namespace std;

/**
 * @class AlignedMmapInstance
 * @brief Memory allocated from here is aligned with respect to Alignment.
 * @author Emery Berger <http://www.cs.umass.edu/~emery>
 */


template <size_t Alignment_>
class AlignedMmapInstance {
public:

  enum { Alignment = Alignment_ };

  inline void * malloc (size_t sz) {

    // Get memory from mmap,
    // align a pointer in the space,
    // and free the space before and after the aligned segment.

    char * ptr = reinterpret_cast<char *>(HL::MmapWrapper::map (sz + Alignment));
    if (ptr == NULL) {
      return NULL;
    }

    char * newptr = align (ptr);

    //
    // Unmap the part before (prolog) and after (prolog).
    //
    // prolog = extra space before the aligned part (PR below).
    // Below is a diagram of the allocated space.
    // p = ptr, n = newptr, EP = epilog.

    // p  n
    // ^  ^-----^  ^
    //  PR  sz   EP

    size_t prolog = (size_t) newptr - (size_t) ptr;

    if (prolog >= Alignment) {
      HL::MmapWrapper::unmap (ptr, prolog);
    }

    // Now unmap the part after.
    // eob = end of buffer (the pointer just after EP).

    void * eob = ptr + Alignment + sz - 1;
    void * originalStart = newptr + sz;
    void * newStart = (void *) align ((char *) originalStart);

    if ((size_t) newStart <= (size_t) eob) {
      size_t epilog = (size_t) eob - (size_t) newStart;
      if (epilog >= Alignment) {
	HL::MmapWrapper::unmap (newStart, epilog);
      }
    }

    // Now record the size (rounded up to the next page) associated
    // with this pointer.

    MyMap.set (newptr, (valType) align ((char *) sz));

    assert (reinterpret_cast<size_t>(newptr) % Alignment == 0);
    return newptr;
  }

  inline void free (void * ptr) {

    // Find the object. If we don't find it, we didn't allocate it.
    // For now, just ignore such an invalid free...

    size_t requestedSize = MyMap.get (ptr);
    if (requestedSize == 0) {
      return;
    }

    HL::MmapWrapper::unmap (ptr, requestedSize);

    // Finally, undo the mapping.
    MyMap.erase (ptr);
  }
  
  inline size_t getSize (void * ptr) {
    return MyMap.get (ptr);
  }


private:

  /// Round a pointer up to the next Alignment-rounded chunk.
  inline static char * align (char * buf) {
    return (char *)(((size_t) buf + (Alignment-1)) & ~(Alignment-1));
  }

  // Manage information in an STL map that uses a custom heap for
  // allocation.

  /// The key is an mmapped pointer.
  typedef const void * keyType;

  /// The value is the requested size.
  typedef size_t valType;

  // The heap from which memory comes for the Map's purposes.
  // Objects come from chunks via mmap, and we manage these with a free list.
  class SourceHeap : public HL::FreelistHeap<BumpAlloc<16384, MmapAlloc> > { };

  /// The map type, with all the pieces in place.
  typedef MyHashMap<keyType, valType, SourceHeap> mapType;

  /// The map that maintains the size of each mmapped chunk.
  mapType MyMap;

};


/**
 * @class AlignedMmap
 * @brief Route requests to the one aligned mmap instance.
 * @author Emery Berger <http://www.cs.umass.edu/~emery>
 */

template <size_t Alignment_>
class AlignedMmap : public ExactlyOne<AlignedMmapInstance<Alignment_> > {
public:

  enum { Alignment = Alignment_ };

  inline void * malloc (size_t sz) {
    return (*this)().malloc (sz);
  }
  inline void free (void * ptr) {
    return (*this)().free (ptr);
  }
  inline size_t getSize (void * ptr) const {
    return (*this)().getSize (ptr);
  }
 
};


#endif
