/*

  Heap Layers: An Extensible Memory Allocation Infrastructure
  
  Copyright (C) 2000-2003 by Emery Berger
  http://www.cs.umass.edu/~emery
  emery@cs.umass.edu
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/**
 * @file   libreap.cpp
 * @brief  Implementations of heap (malloc/free) & reap functions.
 * @author Emery Berger
 */

#ifndef USE_MALLOC_LOCK
#define USE_MALLOC_LOCK 0
#endif

/* pick an allocator implementation (for experimentation only) */

// Reap: the default
#define USE_REAP    1
// Doug Lea's malloc
#define USE_DLHEAP  0
// Windows heap
#define USE_WINHEAP 0

#if defined(_WIN32)
#define _WIN32_WINNT 0x0500
#include <windows.h>
#pragma inline_depth(255)
//#pragma warning(disable: 4530) // exceptions!
#define NO_INLINE __declspec(noinline)

#if !defined(NDEBUG)
#define __forceinline inline
#endif

#else

#define __forceinline inline
#define NO_INLINE

#endif

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

#include <new>

#include "sassert.h"
#include "oneheap.h"
#include "nestedheap.h"
#include "reapletbase.h"
#include "mmapheap.h"
#include "ansiwrapper.h"
#include "seg.h"
#include "adapt.h"
#include "hybridheap.h"
#include "freesllist.h"
#include "combineheap.h"
#include "nullheap.h"
#include "freelistheap.h"
#include "counteddictionary.h"
#include "addreapletheader.h"
#include "lockedheap.h"
#include "spinlock.h"
#include "thinlock.h"

using namespace HL;

#include <iostream>

extern "C" void * dlmalloc (size_t);
extern "C" void dlfree (void *);
extern "C" size_t dlmalloc_usable_size (void *);

extern "C" size_t __cdecl _msize (void *);
extern "C" enum HeapCompatibilityInformation;
extern "C" HANDLE WINAPI GetProcessHeap (void);
extern "C" int WINAPI HeapSetInformation (HANDLE, HEAP_INFORMATION_CLASS, PVOID, SIZE_T);


class LFHMalloc {
public:
  LFHMalloc (void)
  {
    ULONG x=2;
    heap = GetProcessHeap();
    int r = 0;
#if 1
    r = HeapSetInformation(heap,
			   HeapCompatibilityInformation,
			   &x,
			   sizeof(x) );
#endif
    if (r) {
      fprintf (stderr, "Using LFH heap.\n");
    } else {
      fprintf (stderr, "Using standard Windows heap.\n");
    }
  }

  inline void * malloc (size_t sz) {
    return HeapAlloc (heap, 0, sz);
  }
  inline void free (void * ptr) {
    HeapFree (heap, 0, ptr);
  }
  inline size_t getSize (void * ptr) {
    return HeapSize (heap, 0, ptr);
  }
  inline void clear (void) {}
private:
  HANDLE heap;
};

class DLMalloc {
public:
  enum { Alignment = sizeof(double) };

  DLMalloc (void) {
    fprintf (stderr, "Using Doug Lea's malloc.\n");
  }
  __forceinline void * malloc (size_t sz) {
    void * ptr = dlmalloc (sz);
    return ptr;
  }
  __forceinline void free (void * ptr) {
     dlfree (ptr);
    //::free (ptr);
  }
  inline void clear (void) {}
  __forceinline size_t getSize (void * ptr) {
    return dlmalloc_usable_size (ptr);
  }    
};


// allocate reaplets from mmap.
// how do we manage reaplets?

class Bins_4096 {
public:

  enum { NUM_BINS = 92 };
  enum { BIG_OBJECT = 3584 };
  
  static const size_t bins[NUM_BINS];

#if USE_OLD_GETSC
  static inline int getSizeClass (const size_t sz) {
  // Find the size class for a given object size
  // (the smallest i such that _sizeTable[i] >= sz).
    int sizeclass = 0;
    while (bins[sizeclass] < sz)
      sizeclass++;
    return sizeclass;
  }
  
  static inline size_t getClassSize (const int sizeclass) {
  assert (sizeclass >= 0);
  assert (sizeclass < NUM_BINS);
  return bins[sizeclass];
  }

#else
  
  static inline int getSizeClass (size_t sz) {
    assert (sz > 0);
    assert (sz <= BIG_OBJECT);
    int sc;
    --sz;
    if (sz <= 512) {
      sc = sz >> 3;
    } else {
      int sz6 = sz >> 6;
      sc = (sz6 <= 32) ?  56 + sz6 : 85 + (sz >> 9);
    }
    assert (sc >= 0);
    assert (sc < NUM_BINS);    
    return sc;
  }
  
  static inline size_t getClassSize (const int i) {
    assert (i >= 0);
    assert (i < NUM_BINS);
#if 0
    size_t sz;
    if (i < 64) {
      sz = (i+1) << 3;
    } else {
      sz = bins[i];
    }
    return sz;
#else
    return bins[i];
#endif
  }
#endif

};

const size_t Bins_4096::bins[] = {8, 16, 24, 32, 40, 48, 56, 64, 72, 80, 88,
				  96, 104, 112, 120, 128, 136, 144, 152, 160,
				  168, 176, 184, 192, 200, 208, 216, 224, 232,
				  240, 248, 256, 264, 272, 280, 288, 296, 304,
				  312, 320, 328, 336, 344, 352, 360, 368, 376,
				  384, 392, 400, 408, 416, 424, 432, 440, 448,
				  456, 464, 472, 480, 488, 496, 504, 512, 576,
				  640, 704, 768, 832, 896, 960, 1024, 1088, 1152,
				  1216, 1280, 1344, 1408, 1472, 1536, 1600, 1664,
				  1728, 1792, 1856, 1920, 1984, 2048, 2112, 2560,
				  3072, 3584 };




#include "bins16k.h"


class CountingMmapHeap : public MmapHeap {
public:
  CountingMmapHeap (void)
    : _count (0),
      _maxCount (0)
  {}
  ~CountingMmapHeap (void) {
    printf ("count = %d, maxCount = %d\n", _count, _maxCount);
  }
  inline void * malloc (size_t sz) {
    _count += sz;
    if (_count > _maxCount) {
      _maxCount = _count;
    }
    return MmapHeap::malloc (sz);
  }
  inline void free (void * ptr) {
    _count -= getSize(ptr);
    MmapHeap::free (ptr);
  }
private:
  int _count;
  int _maxCount;
};


/* Put in a natural alignment mixin? */

template <class SuperHeap>
class LocalOrMap : public SuperHeap  {
public:
  NO_INLINE void * malloc (size_t sz) {
    void * ptr = SuperHeap::malloc (sz);
    if (!ptr) {
      //          printf ("Foop mmap! malloc %d = %x\n", sz, ptr);
      ptr = MmapHeap::malloc (sz);
      //      printf ("MmapHeap malloc %d = %x\n", sz, ptr);
    }
    return ptr;
  }
};


template <int a, int b>
class Min {
public:
  enum { VALUE = (a < b) ? a : b };
};


//class MainStore : public OneHeap<CountingMmapHeap> {};
class MainStore : public MmapHeap {};

template <int ReapletSize>
class ManageReaplets :
  //  public FreelistHeap<MmapHeap> {
  public LocalOrMap<AdaptHeap<FreeSLList, MmapHeap> > {
  // public Foop<AdaptHeap<CountedDictionary<FreeSLList>, MmapHeap> > {
private:
  class VerifyAlignment :
    public sassert <(Alignment % ReapletSize) == 0> {};
public:
  void clear (void) {}
#if 0
  ManageReaplets (void) {
    fprintf (stderr, "This software uses Reap, version x (prerelease)\nCopyright (C) 2003 Emery Berger\nFor more information, see http://www.cs.umass.edu/~emery.\n");
    //    fprintf (stderr, "woo %d\n", sizeof(ReapletBase));
    //  printf ("S!\n");
  }
#endif
};

#if USE_MALLOC_LOCK && (defined(_MT) || defined(_REENTRANT))
template <int ReapletSize>
class ReapletStore :
  public OneHeap<LockedHeap<SpinLockType, ManageReaplets<ReapletSize> > > {};
#else
template <int ReapletSize>
class ReapletStore : public OneHeap<ManageReaplets<ReapletSize> > {
  //class ReapletStore : public OneHeap<ManageReaplets<ReapletSize> > {
};
#endif

#if 0
class SmallHeap : public Seg<65536, Bins_65536_2, ReapletStore<65536> > {};
class LargeHeap : public AddReapletHeader<65536, MmapHeap> {};
enum { BigSize = 21833 };
#endif

#if 0
//class SmallHeap : public Seg<4096, Bins_4096, ReapletStore<4096> > {};
//class LargeHeap : public AddReapletHeader<4096, MmapHeap> {};
//enum { BigSize = Bins_4096::BIG_OBJECT + 1 };
class SmallHeap : public Seg<32768, Bins_32768_4096, ReapletStore<32768> > {};
class LargeHeap : public AddReapletHeader<32768, MmapHeap> {};
enum { BigSize = 21841 };
#endif

#if 1
class SmallHeap : public Seg<16384, Bins_16384, ReapletStore<16384> > {};
class LargeHeap : public AddReapletHeader<16384, MmapHeap> {};
enum { BigSize = Bins_16384::BIG_OBJECT };
#endif


class NoLockType {
public:
  inline void lock (void) {}
  inline void unlock (void) {}
};


// FIX ME: next step - make sure we're freeing to
// a place under our control (to avoid segfaults).

template <class SuperHeap>
class RobustHeap : public SuperHeap {
public:

  /**
   * @brief Don't free objects that have already been freed.
   * We write a special value into an object's first word to
   * indicate that it's already been freed, counting on the
   * extreme unlikelihood of such a value already sitting in
   * the available space.
   */

  __forceinline void free (void * ptr) {
#if 1
    if (ptr) {

      // FIX ME: we should make sure that these values are not
      // available in ASCII or any reasonable UNICODE string.

      const float alreadyFreedValue = 1.23456789012345678e42;
      float& v = *((float *) ptr);
      if (v != alreadyFreedValue) {
	v = alreadyFreedValue;
	SuperHeap::free (ptr);
      }
    }
#else
    SuperHeap::free (ptr);
#endif
  }
};


#if (USE_REAP + USE_DLHEAP + USE_WINHEAP > 1) || (USE_REAP + USE_DLHEAP + USE_WINHEAP < 1)
#error "Choose exactly one heap implementation."
#endif


#if USE_REAP

class MainHeap :
  public HybridHeap<BigSize, SmallHeap, LargeHeap> {};

class ReapType :
  public NestedHeap<ANSIWrapper<MainHeap> >
{
private:

  // Ensure that this type returns double-word aligned objects.
  class VerifyAlignment :
    public sassert <(Alignment % sizeof(double)) == 0> {};
};

#if USE_MALLOC_LOCK && (defined(_MT) || defined(_REENTRANT))
class TheCustomHeapType : public LockedHeap<SpinLockType, ReapType> {
public:
  NO_INLINE TheCustomHeapType (void) {
    fprintf (stderr, "Using the Reap allocator by Emery Berger (MT version).\n");
  }
};
#else
class TheCustomHeapType : public ReapType {
public:
  NO_INLINE TheCustomHeapType (void) {
    //    fprintf (stderr, "Using the Reap allocator by Emery Berger\n");
  }
};
#endif

#endif


#if USE_DLHEAP
class ReapType : public NestedHeap<DLMalloc> {};
//class ReapType : public NestedHeap<LFHMalloc> {};
class TheCustomHeapType : public ReapType {};
#endif

#if USE_WINHEAP
class ReapType : public NestedHeap<LFHMalloc> {};
class TheCustomHeapType : public ReapType {};
#endif


#if USE_WINHEAP
class ReapType : public NestedHeap<LFHMalloc> {};
class TheCustomHeapType : public ReapType {};
#endif



/// A buffer to hold the custom heap.

static char theCustomHeapBuffer[sizeof(TheCustomHeapType)];

// A static method that guarantees initialization of the custom heap
// and at the same time prevents it from being destroyed (possibly
// prematurely).

__forceinline static TheCustomHeapType * getCustomHeap (void) {
  static TheCustomHeapType * theHeap
    = new ((void *) theCustomHeapBuffer) TheCustomHeapType;
  return theHeap;
}

#include "wrapper.cpp"

extern "C" void reapcreate (void ** reap)
{
  ReapType * r = new ReapType;
  *((ReapType **) reap) = r;
}

extern "C" void * reapmalloc (void ** reap, size_t sz)
{
  return (*((ReapType **) reap))->malloc (sz);
}

extern "C" void reapfree (void ** reap, void * ptr)
{
  return (*((ReapType **) reap))->free (ptr);
}

extern "C" void reapclear (void ** reap)
{
  (*((ReapType **) reap))->clear();
}

extern "C" void reapdestroy (void ** reap)
{
  ReapType * r = *((ReapType **) reap);
  delete r;
  *reap = NULL;
}


extern "C" void regionCreate (void ** reap, void ** parent)
{
  ReapType * r = new ReapType;
  if (parent) {
    (*((ReapType **) parent))->addChild (r);
  }
  *((ReapType **) reap) = r;
}

extern "C" void * regionAllocate (void ** reap, size_t sz)
{
  return (*((ReapType **) reap))->malloc (sz);
}

extern "C" void regionFree (void ** reap, void * ptr)
{
  return (*((ReapType **) reap))->free (ptr);
}

extern "C" void regionFreeAll (void ** reap)
{
  (*((ReapType **) reap))->clear();
}

extern "C" void regionDestroy (void ** reap)
{
  ReapType * r = *((ReapType **) reap);
  //r->clear();
  delete r;
  *reap = NULL;
}


